#include "Fibers.hpp"
#include <new> // For std::nothrow if needed

Fiber::Fiber(void (*function)(), void *data, int priority, FiberState state){
    // Allocate and initialize the stack for the fiber
    data_ = data;
    priority_ = priority;
    state_ = state;

    std::cout << "Current fiber state: " << fiberStateToString(state_) << std::endl;

    stack_bottom = new (std::nothrow) char[4096];
    if (!stack_bottom) {
        throw std::bad_alloc();
    }
    stack_top = stack_bottom + 4096;
    uintptr_t sp = reinterpret_cast<uintptr_t>(stack_top);
    sp = (sp & -16L) - 128; // Align stack to 16 bytes and account for red zone
    context_.rsp = reinterpret_cast<void *>(sp);
    context_.rip = reinterpret_cast<void *>(function);
}

Fiber::~Fiber() {
    delete[] stack_bottom;
}

Context *Fiber::getContext() {
    return &context_;
}

void *Fiber::getData() const {
    return data_;
}

int Fiber::getPriority() const {
    return priority_;
}

void Fiber::setPriority(int priority) {
    priority_ = priority;
}

FiberState Fiber::getState() const {
    return state_;
}

void Fiber::setState(FiberState state) {
    state_ = state;
}
