#ifndef FIBERS_HPP
#define FIBERS_HPP

#include "../Context/context.hpp" // Assuming context.hpp defines the Context struct
#include <cstdint> // For uintptr_t
#include <string>
#include <iostream>

enum class FiberState {
    CREATED,
    READY,
    RUNNING,
    SUSPENDED,
    BLOCKED,
    TERMINATED
};



class Fiber {
private:
    Context context_{};    // Holds the context of the fiber
    void *data_;           // Pointer to data passed to the fiber
    char *stack_bottom;    // Stack memory for the fiber
    char *stack_top;       // Stack memory for the fiber
    int priority_;         // Priority of the fiber (higher value = higher priority)
    FiberState state_;     // Current state of the fiber

public:
    Fiber(void (*function)(), void *data = nullptr, int priority = 0, FiberState state = FiberState::CREATED); // Constructor with priority
    ~Fiber();

    //getters
    Context *getContext();      // Get the context of the fiber
    void *getData() const;      // Get the data passed to the fiber
    int getPriority() const;    // Get the priority of the fiber
    FiberState getState() const; // Get the current state
    
    //setters
    void setState(FiberState state); // Set the current state
    void setPriority(int priority); // Set the priority of the fiber

};

inline std::string fiberStateToString(FiberState state) {
    switch (state) {
        case FiberState::CREATED:    return "CREATED";
        case FiberState::READY:      return "READY";
        case FiberState::RUNNING:    return "RUNNING";
        case FiberState::SUSPENDED:  return "SUSPENDED";
        case FiberState::BLOCKED:    return "BLOCKED";
        case FiberState::TERMINATED: return "TERMINATED";
        default:                     return "UNKNOWN";
    }
}

#endif // FIBERS_HPP
