#include "Scheduler.hpp"

Scheduler::Scheduler() : current_fiber_(nullptr) {}

Scheduler::~Scheduler() {
    while (!fibers_.empty()) {
        delete fibers_.top(); // Delete the highest-priority fiber
        fibers_.pop();
    }
}

void Scheduler::spawn(Fiber *fiber) {
    fibers_.push(fiber); // Add the fiber to the priority queue
    fiber->setState(FiberState::READY);
    std::cout << "Current fiber state: " << fiberStateToString(fiber->getState()) << std::endl;
}

void Scheduler::do_it() {
    get_context(&context_);
    while (!fibers_.empty()) {
        Fiber *current = fibers_.top(); // Get the highest-priority fiber
        fibers_.pop();                  // Remove it from the queue
        current_fiber_ = current;

        current->setState(FiberState::RUNNING);
        std::cout << "Current fiber state: " << fiberStateToString(current->getState()) << std::endl;

        // Switch to the fiber's context
        set_context(current->getContext());
    }
}

void Scheduler::fiber_exit() {
    current_fiber_->setState(FiberState::TERMINATED);
    std::cout << "Current fiber state: " << fiberStateToString(current_fiber_->getState()) << std::endl;
    set_context(&context_);
}

void Scheduler::yield() {
    if (current_fiber_) {
        current_fiber_->setState(FiberState::SUSPENDED);
        std::cout << "Current fiber state: " << fiberStateToString(current_fiber_->getState()) << std::endl;
        // Lower the priority temporarily to put it at the back of the queue
        int original_priority = current_fiber_->getPriority();
        current_fiber_->setPriority(-1); // Set a low priority for this fiber

        fibers_.push(current_fiber_); // Reinsert the current fiber into the queue

        current_fiber_->setState(FiberState::READY);
        std::cout << "Current fiber state: " << fiberStateToString(current_fiber_->getState()) << std::endl;

        // Restore the original priority after reinsertion
        current_fiber_->setPriority(original_priority);


        // Perform the context switch
        swap_context(current_fiber_->getContext(), &context_);
    }
}


void *Scheduler::getData() {
    if (current_fiber_) {
        return current_fiber_->getData();
    }
    return nullptr;
}
