#ifndef SCHEDULER_HPP
#define SCHEDULER_HPP

#include <queue>
#include <iostream>
#include "../Fibers/Fibers.hpp"
#include "../Context/context.hpp" // Assuming context.hpp defines `get_context`, `set_context`, and `swap_context`

// Comparator for priority_queue (higher priority fibers come first)
struct FiberComparator {
    bool operator()(Fiber *f1, Fiber *f2) {
        return f1->getPriority() < f2->getPriority(); // Higher priority comes first
    }
};

class Scheduler {
private:
    std::priority_queue<Fiber *, std::vector<Fiber *>, FiberComparator> fibers_; // Priority queue of fibers
    Context context_; // Context of the scheduler itself
    Fiber *current_fiber_; // Keep track of the currently running fiber

public:
    Scheduler();
    ~Scheduler();

    void spawn(Fiber *fiber); // Add a fiber to the scheduler
    void do_it();             // Run the scheduler
    void fiber_exit();        // Handle fiber exit
    void yield();             // Pause the current fiber and reschedule
    void *getData();          // Get data of the currently running fiber
};

#endif // SCHEDULER_HPP
