#include <iostream>
#include <cstdint>
#include "../context/context.hpp"  // Assuming context.hpp contains declarations for the context functions

//allocateing context
Context foo_context;
Context goo_context;

// Fiber function 'foo' that will be called within a separate context
void foo() {
    std::cout << "You called foo" << std::endl;
    // Simulate fiber exit, we can't return to main as they have different stacks
    set_context(&goo_context);
}

// Fiber function 'goo' that will be called within a separate context
void goo() {
    std::cout << "You called goo" << std::endl;
    // Simulate fiber exit
    exit(1);
}

int main() {
    // Allocate space for two separate stacks: one for 'foo' and one for 'goo'
    char foo_stack[4096];
    char goo_stack[4096];

    // Stack pointers grow downwards, so set the stack pointers to the top of each stack
    uintptr_t sp_foo;
    uintptr_t sp_goo;

    sp_foo = reinterpret_cast<uintptr_t>(foo_stack) + sizeof(foo_stack);
    sp_goo = reinterpret_cast<uintptr_t>(goo_stack) + sizeof(goo_stack);

    // Align the stack pointers to 16 bytes
    sp_foo &= -16L; // this alignment code is now like psudo code please update in read me
    sp_goo &= -16L;   

    sp_foo -= 128;
    sp_goo -= 128;

    // Set the instruction pointer (rip) and stack pointer (rsp) for both fibers
    foo_context.rip = reinterpret_cast<void*>(foo);
    foo_context.rsp = reinterpret_cast<void*>(sp_foo);  // Account for the red zone

    goo_context.rip = reinterpret_cast<void*>(goo);
    goo_context.rsp = reinterpret_cast<void*>(sp_goo);  // Account for the red zone

    // Output before setting contexts to show control will transfer to fibers
    std::cout << "Main function, calling foo and goo" << std::endl;

    // Set context for 'foo', this will transfer control to 'foo' on execution
    set_context(&foo_context);


    return 0;
}
