#include <iostream>
#include <cstdint>
#include "../context/context.hpp"  // Assuming context.hpp contains declarations for the context functions


// Fiber class definition
class Fiber {
private:
    Context context_{}; // Holds the context of the fiber
    char *stack_bottom; // Stack memory for the fiber
    char *stack_top; // Stack memory for the fiber

public:
    Fiber(void (*function)()) {
        // Allocate and initialize the stack for the fiber
        stack_bottom = new char[4096];
        stack_top = stack_bottom + 4096;
        uintptr_t sp = reinterpret_cast<uintptr_t>(stack_top);
        sp = (sp & -16L) - 128; // Align stack to 16 bytes and account for red zone
        context_.rsp = reinterpret_cast<void *>(sp);
        context_.rip = reinterpret_cast<void *>(function);
    }

    ~Fiber() {
        delete[] stack_bottom;
    }

    Context *getContext() { return &context_; }

};


// Fiber function 'foo' that will be called within a separate context
void foo() {
    std::cout << "You called foo" << std::endl;
    // Simulate fiber exit, we can't return to main as they have different stacks
    exit(1);
}



int main() {


    // Output before setting contexts to show control will transfer to fibers
    std::cout << "Main function, calling foo" << std::endl;

    Fiber foo_fiber(foo);
    // Set context for 'foo', this will transfer control to 'foo' on execution
    set_context(foo_fiber.getContext());

    return 0;
}
