#include "Scheduler.hpp"

Scheduler::Scheduler() : current_fiber_(nullptr) {}

Scheduler::~Scheduler() {
    while (!fibers_.empty()) {
        delete fibers_.front();
        fibers_.pop_front();
    }
}

void Scheduler::spawn(Fiber *fiber) {
    fibers_.push_back(fiber);
}

void Scheduler::do_it() {
    get_context(&context_);
    while (!fibers_.empty()) {
        Fiber *current = fibers_.front();
        fibers_.pop_front();
        current_fiber_ = current;

        // Switch to the fiber's context
        set_context(current->getContext());
    }
}

void Scheduler::fiber_exit() {
    set_context(&context_);
}

void Scheduler::yield() {
    if (current_fiber_) {
        fibers_.push_back(current_fiber_);
        swap_context(current_fiber_->getContext(), &context_);
    }
}

void *Scheduler::getData() {
    if (current_fiber_) {
        return current_fiber_->getData();
    }
    return nullptr;
}
